#--------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved. 
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#--------------------------------------------------------------------------

import unittest
import json
try:
    from unittest import mock
except ImportError:
    import mock

import requests
import httpretty

from msrestazure.azure_configuration import AzureConfiguration
from msrestazure.tools import parse_resource_id, is_valid_resource_id, resource_id

class TestTools(unittest.TestCase):

    @httpretty.activate
    @mock.patch('time.sleep', return_value=None)
    def test_register_rp_hook(self, time_sleep):
        """Protocol:
        - We call the provider and get a 409 provider error
        - Now we POST register provider and get "Registering"
        - Now we GET register provider and get "Registered"
        - We call again the first endpoint and this time this succeed
        """

        provider_url = ("https://management.azure.com/"
                        "subscriptions/12345678-9abc-def0-0000-000000000000/"
                        "resourceGroups/clitest.rg000001/"
                        "providers/Microsoft.Sql/servers/ygserver123?api-version=2014-04-01")

        provider_error = ('{"error":{"code":"MissingSubscriptionRegistration", '
                          '"message":"The subscription registration is in \'Unregistered\' state. '
                          'The subscription must be registered to use namespace \'Microsoft.Sql\'. '
                          'See https://aka.ms/rps-not-found for how to register subscriptions."}}')

        provider_success = '{"success": true}'

        httpretty.register_uri(httpretty.PUT,
                               provider_url,
                               responses=[
                                   httpretty.Response(body=provider_error, status=409),
                                   httpretty.Response(body=provider_success),
                               ],
                               content_type="application/json")

        register_post_url = ("https://management.azure.com/"
                             "subscriptions/12345678-9abc-def0-0000-000000000000/"
                             "providers/Microsoft.Sql/register?api-version=2016-02-01")

        register_post_result = {
            "registrationState":"Registering"
        }

        register_get_url = ("https://management.azure.com/"
                            "subscriptions/12345678-9abc-def0-0000-000000000000/"
                            "providers/Microsoft.Sql?api-version=2016-02-01")

        register_get_result = {
            "registrationState":"Registered"
        }

        httpretty.register_uri(httpretty.POST,
                               register_post_url,
                               body=json.dumps(register_post_result),
                               content_type="application/json")

        httpretty.register_uri(httpretty.GET,
                               register_get_url,
                               body=json.dumps(register_get_result),
                               content_type="application/json")

        configuration = AzureConfiguration(None)
        register_rp_hook = configuration.hooks[0]

        session = requests.Session()
        def rp_cb(r, *args, **kwargs):
            kwargs.setdefault("msrest", {})["session"] = session
            return register_rp_hook(r, *args, **kwargs)
        session.hooks['response'].append(rp_cb)

        response = session.put(provider_url)
        self.assertTrue(response.json()['success'])

    @httpretty.activate
    @mock.patch('time.sleep', return_value=None)
    def test_register_failed(self, time_sleep):
        """Protocol:
        - We call the provider and get a 409 provider error
        - Now we POST register provider and get "Registering"
        - This POST failed
        """

        provider_url = ("https://management.azure.com/"
                        "subscriptions/12345678-9abc-def0-0000-000000000000/"
                        "resourceGroups/clitest.rg000001/"
                        "providers/Microsoft.Sql/servers/ygserver123?api-version=2014-04-01")

        provider_error = ('{"error":{"code":"MissingSubscriptionRegistration", '
                          '"message":"The subscription registration is in \'Unregistered\' state. '
                          'The subscription must be registered to use namespace \'Microsoft.Sql\'. '
                          'See https://aka.ms/rps-not-found for how to register subscriptions."}}')

        provider_success = '{"success": true}'

        httpretty.register_uri(httpretty.PUT,
                               provider_url,
                               responses=[
                                   httpretty.Response(body=provider_error, status=409),
                                   httpretty.Response(body=provider_success),
                               ],
                               content_type="application/json")

        register_post_url = ("https://management.azure.com/"
                             "subscriptions/12345678-9abc-def0-0000-000000000000/"
                             "providers/Microsoft.Sql/register?api-version=2016-02-01")

        httpretty.register_uri(httpretty.POST,
                               register_post_url,
                               status=409,
                               content_type="application/json")

        configuration = AzureConfiguration(None)
        register_rp_hook = configuration.hooks[0]

        session = requests.Session()
        def rp_cb(r, *args, **kwargs):
            kwargs.setdefault("msrest", {})["session"] = session
            return register_rp_hook(r, *args, **kwargs)
        session.hooks['response'].append(rp_cb)

        response = session.put(provider_url)
        self.assertEqual(409, response.status_code)

    def test_resource_parse(self):
        """
        Tests resource id parsing, reforming, and validation.
        """
        tests = [
            {
                'resource_id': '/subscriptions/fakesub/resourcegroups/testgroup/providers'
                               '/Microsoft.Storage/storageAccounts/foo/providers'
                               '/Microsoft.Authorization/locks/bar',
                'expected': {
                    'name': 'foo',
                    'type': 'storageAccounts',
                    'namespace': 'Microsoft.Storage',
                    'child_name_1': 'bar',
                    'child_namespace_1': 'Microsoft.Authorization',
                    'child_type_1': 'locks',
                    'child_parent_1': 'storageAccounts/foo/providers/Microsoft.Authorization/',
                    'resource_group': 'testgroup',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/subscriptions/fakesub/resourcegroups/testgroup/providers'
                               '/Microsoft.Storage/storageAccounts/foo'
                               '/locks/bar',
                'expected': {
                    'name': 'foo',
                    'type': 'storageAccounts',
                    'namespace': 'Microsoft.Storage',
                    'child_name_1': 'bar',
                    'child_type_1': 'locks',
                    'child_parent_1': 'storageAccounts/foo/',
                    'resource_group': 'testgroup',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/subscriptions/fakesub/resourcegroups/testgroup/providers'
                               '/Microsoft.Storage/storageAccounts/foo/providers'
                               '/Microsoft.Authorization/locks/bar/providers/Microsoft.Network/'
                               'nets/gc',
                'expected': {
                    'name': 'foo',
                    'type': 'storageAccounts',
                    'namespace': 'Microsoft.Storage',
                    'child_name_1': 'bar',
                    'child_namespace_1': 'Microsoft.Authorization',
                    'child_type_1': 'locks',
                    'child_parent_1': 'storageAccounts/foo/providers/Microsoft.Authorization/',
                    'child_name_2': 'gc',
                    'child_namespace_2': 'Microsoft.Network',
                    'child_type_2': 'nets',
                    'child_parent_2': 'storageAccounts/foo/providers/Microsoft.Authorization/'
                                      'locks/bar/providers/Microsoft.Network/',
                    'resource_group': 'testgroup',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/subscriptions/fakesub/resourcegroups/testgroup/providers'
                               '/Microsoft.Storage/storageAccounts/foo'
                               '/locks/bar/nets/gc',
                'expected': {
                    'name': 'foo',
                    'type': 'storageAccounts',
                    'namespace': 'Microsoft.Storage',
                    'child_name_1': 'bar',
                    'child_type_1': 'locks',
                    'child_parent_1': 'storageAccounts/foo/',
                    'child_name_2': 'gc',
                    'child_type_2': 'nets',
                    'child_parent_2': 'storageAccounts/foo/locks/bar/',
                    'resource_group': 'testgroup',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg/providers/'
                               'Microsoft.Provider1/resourceType1/name1',
                'expected': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.Provider1',
                    'type': 'resourceType1',
                    'name': 'name1',
                    'resource_parent': '',
                    'resource_namespace': 'Microsoft.Provider1',
                    'resource_type': 'resourceType1',
                    'resource_name': 'name1'
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg/providers/'
                               'Microsoft.Provider1/resourceType1/name1/resourceType2/name2',
                'expected': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.Provider1',
                    'type': 'resourceType1',
                    'name': 'name1',
                    'child_namespace_1': None,
                    'child_type_1': 'resourceType2',
                    'child_name_1': 'name2',
                    'child_parent_1': 'resourceType1/name1/',
                    'resource_parent': 'resourceType1/name1/',
                    'resource_namespace': 'Microsoft.Provider1',
                    'resource_type': 'resourceType2',
                    'resource_name': 'name2'
                }
            },
            {
                'resource_id': '/subscriptions/00000/resourceGroups/myRg/providers/'
                               'Microsoft.RecoveryServices/vaults/vault_name/backupFabrics/'
                               'fabric_name/protectionContainers/container_name/'
                               'protectedItems/item_name/recoveryPoint/recovery_point_guid',
                'expected': {
                    'subscription': '00000',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.RecoveryServices',
                    'type': 'vaults',
                    'name': 'vault_name',
                    'child_type_1': 'backupFabrics',
                    'child_name_1': 'fabric_name',
                    'child_parent_1': 'vaults/vault_name/',
                    'child_type_2': 'protectionContainers',
                    'child_name_2': 'container_name',
                    'child_parent_2': 'vaults/vault_name/backupFabrics/fabric_name/',
                    'child_type_3': 'protectedItems',
                    'child_name_3': 'item_name',
                    'child_parent_3': 'vaults/vault_name/backupFabrics/fabric_name/'
                                      'protectionContainers/container_name/',
                    'child_type_4': 'recoveryPoint',
                    'child_name_4': 'recovery_point_guid',
                    'child_parent_4': 'vaults/vault_name/backupFabrics/fabric_name/'
                                      'protectionContainers/container_name/protectedItems/'
                                      'item_name/',
                    'resource_parent': 'vaults/vault_name/backupFabrics/fabric_name/'
                                       'protectionContainers/container_name/protectedItems/'
                                       'item_name/',
                    'resource_namespace': 'Microsoft.RecoveryServices',
                    'resource_type': 'recoveryPoint',
                    'resource_name': 'recovery_point_guid'
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg/providers/'
                               'Microsoft.Provider1/resourceType1/name1/resourceType2/name2/'
                               'providers/Microsoft.Provider3/resourceType3/name3',
                'expected': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.Provider1',
                    'type': 'resourceType1',
                    'name': 'name1',
                    'child_namespace_1': None,
                    'child_type_1': 'resourceType2',
                    'child_name_1': 'name2',
                    'child_parent_1' : 'resourceType1/name1/',
                    'child_namespace_2': 'Microsoft.Provider3',
                    'child_type_2': 'resourceType3',
                    'child_name_2': 'name3',
                    'child_parent_2': 'resourceType1/name1/resourceType2/name2/'
                                      'providers/Microsoft.Provider3/',
                    'resource_parent': 'resourceType1/name1/resourceType2/name2/'
                                       'providers/Microsoft.Provider3/',
                    'resource_namespace': 'Microsoft.Provider1',
                    'resource_type': 'resourceType3',
                    'resource_name': 'name3'
                }
            },
            {
                'resource_id': '/subscriptions/fakesub/providers/Microsoft.Authorization'
                               '/locks/foo',
                'expected': {
                    'name': 'foo',
                    'type': 'locks',
                    'namespace': 'Microsoft.Authorization',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/Subscriptions/fakesub/providers/Microsoft.Authorization'
                               '/locks/foo',
                'expected': {
                    'name': 'foo',
                    'type': 'locks',
                    'namespace': 'Microsoft.Authorization',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg',
                'expected': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg'
                }
            }
        ]
        for test in tests:
            self.assertTrue(is_valid_resource_id(test['resource_id']))
            kwargs = parse_resource_id(test['resource_id'])
            for key in test['expected']:
                try:
                    self.assertEqual(kwargs[key], test['expected'][key])
                except KeyError:
                    self.assertTrue(key not in kwargs and test['expected'][key] is None)

        invalid_ids = [
            '/subscriptions/fakesub/resourceGroups/myRg/type1/name1',
            '/subscriptions/fakesub/resourceGroups/myRg/providers/Microsoft.Provider/foo',
            '/subscriptions/fakesub/resourceGroups/myRg/providers/namespace/type/name/type1'
        ]
        for invalid_id in invalid_ids:
            self.assertFalse(is_valid_resource_id(invalid_id))

        tests = [
            {
                'resource_id': '/subscriptions/fakesub/resourcegroups/testgroup/providers'
                               '/Microsoft.Storage/storageAccounts/foo/providers'
                               '/Microsoft.Authorization/locks/bar',
                'id_args': {
                    'name': 'foo',
                    'type': 'storageAccounts',
                    'namespace': 'Microsoft.Storage',
                    'child_name_1': 'bar',
                    'child_namespace_1': 'Microsoft.Authorization',
                    'child_type_1': 'locks',
                    'resource_group': 'testgroup',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/subscriptions/fakesub/resourcegroups/testgroup/providers'
                               '/Microsoft.Storage/storageAccounts/foo'
                               '/locks/bar',
                'id_args': {
                    'name': 'foo',
                    'type': 'storageAccounts',
                    'namespace': 'Microsoft.Storage',
                    'child_name_1': 'bar',
                    'child_type_1': 'locks',
                    'resource_group': 'testgroup',
                    'subscription': 'fakesub',
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg/providers/'
                               'Microsoft.Provider1/resourceType1/name1/resourceType2/name2/'
                               'providers/Microsoft.Provider3/resourceType3/name3',
                'id_args': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.Provider1',
                    'type': 'resourceType1',
                    'name': 'name1',
                    'child_type_1': 'resourceType2',
                    'child_name_1': 'name2',
                    'child_namespace_2': 'Microsoft.Provider3',
                    'child_type_2': 'resourceType3',
                    'child_name_2': 'name3'
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg/'
                               'providers/Microsoft.Provider1',
                'id_args': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.Provider1'
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg',
                'id_args': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg'
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg/'
                               'providers/Microsoft.Provider1/resourceType1/name1/resourceType2/'
                               'name2/providers/Microsoft.Provider3',
                'id_args': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.Provider1',
                    'type': 'resourceType1',
                    'name': 'name1',
                    'child_type_1': 'resourceType2',
                    'child_name_1': 'name2',
                    'child_namespace_2': 'Microsoft.Provider3'
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg/'
                               'providers/Microsoft.Provider1/resourceType1/name1',
                'id_args': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg',
                    'namespace': 'Microsoft.Provider1',
                    'type': 'resourceType1',
                    'name': 'name1',
                    'child_type_1': None,
                    'child_name_1': 'name2',
                    'child_namespace_2': 'Microsoft.Provider3'
                }
            },
            {
                'resource_id': '/subscriptions/mySub/resourceGroups/myRg',
                'id_args': {
                    'subscription': 'mySub',
                    'resource_group': 'myRg'
                }
            }
        ]
        for test in tests:
            rsrc_id = resource_id(**test['id_args'])
            self.assertEqual(rsrc_id.lower(), test['resource_id'].lower())


if __name__ == "__main__":
    unittest.main()



#--------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved. 
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#--------------------------------------------------------------------------

import unittest
import json
import httpretty
import requests

from msrestazure import azure_cloud

class TestCloud(unittest.TestCase):

    @httpretty.activate
    def test_get_cloud_from_endpoint(self):

        public_azure_dict = {
        	"galleryEndpoint": "https://gallery.azure.com",
	        "graphEndpoint": "https://graph.windows.net/",
	        "portalEndpoint": "https://portal.azure.com",
	        "authentication": {
		        "loginEndpoint": "https://login.windows.net",
		        "audiences": ["https://management.core.windows.net/", "https://management.azure.com/"]
	        }
        }

        httpretty.register_uri(httpretty.GET,
                               "https://management.azure.com/metadata/endpoints?api-version=1.0",
                               body=json.dumps(public_azure_dict),
                               content_type="application/json")

        cloud = azure_cloud.get_cloud_from_metadata_endpoint("https://management.azure.com")
        self.assertEqual("https://management.azure.com", cloud.name)
        self.assertEqual("https://management.azure.com", cloud.endpoints.management)
        self.assertEqual("https://management.azure.com", cloud.endpoints.resource_manager)
        self.assertEqual("https://gallery.azure.com", cloud.endpoints.gallery)
        self.assertEqual("https://graph.windows.net/", cloud.endpoints.active_directory_graph_resource_id)
        self.assertEqual("https://login.windows.net", cloud.endpoints.active_directory)

        session = requests.Session()
        cloud = azure_cloud.get_cloud_from_metadata_endpoint("https://management.azure.com", "Public Azure", session)
        self.assertEqual("Public Azure", cloud.name)
        self.assertEqual("https://management.azure.com", cloud.endpoints.management)
        self.assertEqual("https://management.azure.com", cloud.endpoints.resource_manager)
        self.assertEqual("https://gallery.azure.com", cloud.endpoints.gallery)
        self.assertEqual("https://graph.windows.net/", cloud.endpoints.active_directory_graph_resource_id)
        self.assertEqual("https://login.windows.net", cloud.endpoints.active_directory)

        with self.assertRaises(azure_cloud.MetadataEndpointError):
            azure_cloud.get_cloud_from_metadata_endpoint("https://something.azure.com")

        with self.assertRaises(azure_cloud.CloudEndpointNotSetException):
            cloud.endpoints.batch_resource_id

        with self.assertRaises(azure_cloud.CloudSuffixNotSetException):
            cloud.suffixes.sql_server_hostname

        self.assertIsNotNone(str(cloud))



#--------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#--------------------------------------------------------------------------

import json
import re
import types
import unittest
try:
    from unittest import mock
except ImportError:
    import mock

import httpretty
import pytest

from requests import Request, Response

from msrest import Deserializer, Configuration
from msrest.service_client import ServiceClient
from msrest.exceptions import DeserializationError
from msrest.polling import LROPoller
from msrestazure.azure_exceptions import CloudError
from msrestazure.polling.arm_polling import (
    LongRunningOperation,
    ARMPolling,
    BadStatus
)

class SimpleResource:
    """An implementation of Python 3 SimpleNamespace.
    Used to deserialize resource objects from response bodies where
    no particular object type has been specified.
    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        keys = sorted(self.__dict__)
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

class BadEndpointError(Exception):
    pass

TEST_NAME = 'foo'
RESPONSE_BODY = {'properties':{'provisioningState': 'InProgress'}}
ASYNC_BODY = json.dumps({ 'status': 'Succeeded' })
ASYNC_URL = 'http://dummyurlFromAzureAsyncOPHeader_Return200'
LOCATION_BODY = json.dumps({ 'name': TEST_NAME })
LOCATION_URL = 'http://dummyurlurlFromLocationHeader_Return200'
RESOURCE_BODY = json.dumps({ 'name': TEST_NAME })
RESOURCE_URL = 'http://subscriptions/sub1/resourcegroups/g1/resourcetype1/resource1'
ERROR = 'http://dummyurl_ReturnError'
POLLING_STATUS = 200

CLIENT = ServiceClient(None, Configuration("http://example.org"))
def mock_send(client_self, request, header_parameters, stream):
    return TestArmPolling.mock_update(request.url, header_parameters)
CLIENT.send = types.MethodType(mock_send, CLIENT)


class TestArmPolling(object):

    convert = re.compile('([a-z0-9])([A-Z])')

    @staticmethod
    def mock_send(method, status, headers=None, body=None):
        if headers is None:
            headers = {}
        response = mock.create_autospec(Response)
        response.request = mock.create_autospec(Request)
        response.request.method = method
        response.request.url = RESOURCE_URL
        response.request.headers = {
            'x-ms-client-request-id': '67f4dd4e-6262-45e1-8bed-5c45cf23b6d9'
        }
        response.status_code = status
        response.headers = headers
        response.headers.update({"content-type": "application/json; charset=utf8"})
        content = body if body is not None else RESPONSE_BODY
        response.text = json.dumps(content)
        response.json = lambda: json.loads(response.text)
        return response

    @staticmethod
    def mock_update(url, headers=None):
        response = mock.create_autospec(Response)
        response.request = mock.create_autospec(Request)
        response.request.method = 'GET'
        response.headers = headers or {}
        response.headers.update({"content-type": "application/json; charset=utf8"})

        if url == ASYNC_URL:
            response.request.url = url
            response.status_code = POLLING_STATUS
            response.text = ASYNC_BODY
            response.randomFieldFromPollAsyncOpHeader = None

        elif url == LOCATION_URL:
            response.request.url = url
            response.status_code = POLLING_STATUS
            response.text = LOCATION_BODY
            response.randomFieldFromPollLocationHeader = None

        elif url == ERROR:
            raise BadEndpointError("boom")

        elif url == RESOURCE_URL:
            response.request.url = url
            response.status_code = POLLING_STATUS
            response.text = RESOURCE_BODY

        else:
            raise Exception('URL does not match')
        response.json = lambda: json.loads(response.text)
        return response

    @staticmethod
    def mock_outputs(response):
        body = response.json()
        body = {TestArmPolling.convert.sub(r'\1_\2', k).lower(): v
                for k, v in body.items()}
        properties = body.setdefault('properties', {})
        if 'name' in body:
            properties['name'] = body['name']
        if properties:
            properties = {TestArmPolling.convert.sub(r'\1_\2', k).lower(): v
                          for k, v in properties.items()}
            del body['properties']
            body.update(properties)
            resource = SimpleResource(**body)
        else:
            raise DeserializationError("Impossible to deserialize")
            resource = SimpleResource(**body)
        return resource

    def test_long_running_put(self):
        #TODO: Test custom header field

        # Test throw on non LRO related status code
        response = TestArmPolling.mock_send('PUT', 1000, {})
        op = LongRunningOperation(response, lambda x:None)
        with pytest.raises(BadStatus):
            op.set_initial_status(response)
        with pytest.raises(CloudError):
            LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

        # Test with no polling necessary
        response_body = {
            'properties':{'provisioningState': 'Succeeded'},
            'name': TEST_NAME
        }
        response = TestArmPolling.mock_send(
            'PUT', 201,
            {}, response_body
        )
        def no_update_allowed(url, headers=None):
            raise ValueError("Should not try to update")
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0)
        )
        assert poll.result().name == TEST_NAME
        assert not hasattr(poll._polling_method._response, 'randomFieldFromPollAsyncOpHeader')

        # Test polling from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'PUT', 201,
            {'azure-asyncoperation': ASYNC_URL})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert not hasattr(poll._polling_method._response, 'randomFieldFromPollAsyncOpHeader')

        # Test polling location header
        response = TestArmPolling.mock_send(
            'PUT', 201,
            {'location': LOCATION_URL})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert poll._polling_method._response.randomFieldFromPollLocationHeader is None

        # Test polling initial payload invalid (SQLDb)
        response_body = {}  # Empty will raise
        response = TestArmPolling.mock_send(
            'PUT', 201,
            {'location': LOCATION_URL}, response_body)
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert poll._polling_method._response.randomFieldFromPollLocationHeader is None

        # Test fail to poll from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'PUT', 201,
            {'azure-asyncoperation': ERROR})
        with pytest.raises(BadEndpointError):
            poll = LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

        # Test fail to poll from location header
        response = TestArmPolling.mock_send(
            'PUT', 201,
            {'location': ERROR})
        with pytest.raises(BadEndpointError):
            poll = LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

    def test_long_running_patch(self):

        # Test polling from location header
        response = TestArmPolling.mock_send(
            'PATCH', 202,
            {'location': LOCATION_URL},
            body={'properties':{'provisioningState': 'Succeeded'}})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert poll._polling_method._response.randomFieldFromPollLocationHeader is None

        # Test polling from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'PATCH', 202,
            {'azure-asyncoperation': ASYNC_URL},
            body={'properties':{'provisioningState': 'Succeeded'}})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert not hasattr(poll._polling_method._response, 'randomFieldFromPollAsyncOpHeader')

        # Test polling from location header
        response = TestArmPolling.mock_send(
            'PATCH', 200,
            {'location': LOCATION_URL},
            body={'properties':{'provisioningState': 'Succeeded'}})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert poll._polling_method._response.randomFieldFromPollLocationHeader is None

        # Test polling from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'PATCH', 200,
            {'azure-asyncoperation': ASYNC_URL},
            body={'properties':{'provisioningState': 'Succeeded'}})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert not hasattr(poll._polling_method._response, 'randomFieldFromPollAsyncOpHeader')

        # Test fail to poll from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'PATCH', 202,
            {'azure-asyncoperation': ERROR})
        with pytest.raises(BadEndpointError):
            poll = LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

        # Test fail to poll from location header
        response = TestArmPolling.mock_send(
            'PATCH', 202,
            {'location': ERROR})
        with pytest.raises(BadEndpointError):
            poll = LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

    def test_long_running_delete(self):
        # Test polling from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'DELETE', 202,
            {'azure-asyncoperation': ASYNC_URL},
            body=""
        )
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        poll.wait()
        assert poll.result() is None
        assert poll._polling_method._response.randomFieldFromPollAsyncOpHeader is None

    @httpretty.activate
    def test_long_running_post(self):

        # Test POST LRO with both Location and Azure-AsyncOperation

        # The initial response contains both Location and Azure-AsyncOperation, a 202 and no Body
        response = TestArmPolling.mock_send(
            'POST',
            202,
            {
                'location': 'http://example.org/location',
                'azure-asyncoperation': 'http://example.org/async_monitor',
            },
            ''
        )

        class TestServiceClient(ServiceClient):
            def __init__(self):
                ServiceClient.__init__(self, None, Configuration("http://example.org"))

            def send(self, request, headers=None, content=None, **config):
                assert request.method == 'GET'

                if request.url == 'http://example.org/location':
                    return TestArmPolling.mock_send(
                        'GET',
                        200,
                        body={'location_result': True}
                    )
                elif request.url == 'http://example.org/async_monitor':
                    return TestArmPolling.mock_send(
                        'GET',
                        200,
                        body={'status': 'Succeeded'}
                    )
                else:
                    pytest.fail("No other query allowed")

        def deserialization_cb(response):
            return response.json()

        # Test 1, LRO options with Location final state
        poll = LROPoller(
            TestServiceClient(),
            response,
            deserialization_cb,
            ARMPolling(0, lro_options={"final-state-via": "location"}))
        result = poll.result()
        assert result['location_result'] == True

        # Test 2, LRO options with Azure-AsyncOperation final state
        poll = LROPoller(
            TestServiceClient(),
            response,
            deserialization_cb,
            ARMPolling(0, lro_options={"final-state-via": "azure-async-operation"}))
        result = poll.result()
        assert result['status'] == 'Succeeded'

        # Test 3, backward compat (no options, means "azure-async-operation")
        poll = LROPoller(
            TestServiceClient(),
            response,
            deserialization_cb,
            ARMPolling(0))
        result = poll.result()
        assert result['status'] == 'Succeeded'

        # Test 4, location has no body

        class TestServiceClientNoBody(ServiceClient):
            def __init__(self):
                ServiceClient.__init__(self, None, Configuration("http://example.org"))

            def send(self, request, headers=None, content=None, **config):
                assert request.method == 'GET'

                if request.url == 'http://example.org/location':
                    return TestArmPolling.mock_send(
                        'GET',
                        200,
                        body=""
                    )
                elif request.url == 'http://example.org/async_monitor':
                    return TestArmPolling.mock_send(
                        'GET',
                        200,
                        body={'status': 'Succeeded'}
                    )
                else:
                    pytest.fail("No other query allowed")

        poll = LROPoller(
            TestServiceClientNoBody(),
            response,
            deserialization_cb,
            ARMPolling(0, lro_options={"final-state-via": "location"}))
        result = poll.result()
        assert result is None


        # Former oooooold tests to refactor one day to something more readble

        # Test throw on non LRO related status code
        response = TestArmPolling.mock_send('POST', 201, {})
        op = LongRunningOperation(response, lambda x:None)
        with pytest.raises(BadStatus):
            op.set_initial_status(response)
        with pytest.raises(CloudError):
            LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

        # Test polling from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'POST', 202,
            {'azure-asyncoperation': ASYNC_URL},
            body={'properties':{'provisioningState': 'Succeeded'}})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        poll.wait()
        #self.assertIsNone(poll.result())
        assert poll._polling_method._response.randomFieldFromPollAsyncOpHeader is None

        # Test polling from location header
        response = TestArmPolling.mock_send(
            'POST', 202,
            {'location': LOCATION_URL},
            body={'properties':{'provisioningState': 'Succeeded'}})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        assert poll.result().name == TEST_NAME
        assert poll._polling_method._response.randomFieldFromPollLocationHeader is None

        # Test fail to poll from azure-asyncoperation header
        response = TestArmPolling.mock_send(
            'POST', 202,
            {'azure-asyncoperation': ERROR})
        with pytest.raises(BadEndpointError):
            poll = LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

        # Test fail to poll from location header
        response = TestArmPolling.mock_send(
            'POST', 202,
            {'location': ERROR})
        with pytest.raises(BadEndpointError):
            poll = LROPoller(CLIENT, response,
                TestArmPolling.mock_outputs,
                ARMPolling(0)).result()

    def test_long_running_negative(self):
        global LOCATION_BODY
        global POLLING_STATUS

        # Test LRO PUT throws for invalid json
        LOCATION_BODY = '{'
        response = TestArmPolling.mock_send(
            'POST', 202,
            {'location': LOCATION_URL})
        poll = LROPoller(
            CLIENT,
            response,
            TestArmPolling.mock_outputs,
            ARMPolling(0)
        )
        with pytest.raises(DeserializationError):
            poll.wait()

        LOCATION_BODY = '{\'"}'
        response = TestArmPolling.mock_send(
            'POST', 202,
            {'location': LOCATION_URL})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        with pytest.raises(DeserializationError):
            poll.wait()

        LOCATION_BODY = '{'
        POLLING_STATUS = 203
        response = TestArmPolling.mock_send(
            'POST', 202,
            {'location': LOCATION_URL})
        poll = LROPoller(CLIENT, response,
            TestArmPolling.mock_outputs,
            ARMPolling(0))
        with pytest.raises(CloudError): # TODO: Node.js raises on deserialization
            poll.wait()

        LOCATION_BODY = json.dumps({ 'name': TEST_NAME })
        POLLING_STATUS = 200

